/* CS184 - Spring 2003 - Project
 *
 * Name: Ben Young
 * Login: cs194-fx
 */

#include <time.h>
#include <stdlib.h>
#include "Landscape.h"
#include "LandscapeTriangle.h"
#include "Camera.h"
#include "Vertex.h"
#include "Particle.h"
#include "Matrix2D.h"
#include "Vector2D.h"
#include "Matrix.h"


/* Size of our window */
#define RES_X 640
#define RES_Y 480

/* Near/Far Clipping */
#define PNEAR 0.0001
#define PFAR  1000000

/* Field of view */
#define FOVY 60.0

/* Global variables to save global state */

const static GLfloat KB_INTERVAL = .05;


static LandscapeTriangle * terrain;

static Camera *camera = NULL;

static GLint lastx;
static GLint lasty;

static GLUquadric * quad = gluNewQuadric();


static Vector *trans = new Vector(0,0,0,0);

static Matrix2D terrainTransform;

static list<Particle*> * particleList = new list<Particle*>();
static hash_map<int, hash_map<int, Particle*> * > * ht1 = new hash_map<int, hash_map<int, Particle*> * >();
static hash_map<int, hash_map<int, Particle*> * > * ht2 = new hash_map<int, hash_map<int, Particle*> * >();

/* Do we fill our polygons */
static bool fillPolygons = true;
static bool locked = false;

static bool depthTest = true;
static bool rain = false;

static aDown;
static sDown;
static wDown;
static dDown;
static qDown;
static eDown;


static void resetCamera() {
	if(camera==NULL) {
		camera = new Camera();
	}
	camera->eye  = Vector(0,.5,-1, 1);
	camera->look = Vector(0, 0, 1, 0);
	camera->up   = Vector(0, 1, 0, 0);
}


static void killMotion() {
	if(!(qDown || wDown || eDown || aDown || sDown || dDown)) {
		*trans=Vector();
	}
}


/* The display function. It is called whenever the window needs
 * redrawing (ie: overlapping window moves, resize, maximize, or 
 * glutPostRedisplay() is called)  You should draw your polygons here.
 */
static void display() {



	//unsigned int i;
	//POLYGON_LIST *polys;
	//POINT_LIST *points;

	//list <Polygon <GLfloat>> polys;
	//list <Vector <GLfloat>> points;
	Vector * n=NULL;
	GLfloat oldHeight;
	GLfloat dh = 0;
	if(terrain->hit(camera->eye.x, camera->eye.z, &oldHeight, &n)) {
		GLfloat newHeight;
		if(terrain->hit(camera->eye.x + trans->x, camera->eye.z + trans->z, &newHeight, &n)) {
			if(camera->eye.y < oldHeight + 1) {
				dh =  newHeight - oldHeight;
			}
		}
	}
	Vector hover = Vector(trans->x, trans->y + dh, trans->z, 1);
	
	

	*camera = camera->translate(hover);

	

	// Enable depth testing for hidden line removal
	if(depthTest) {
		glEnable(GL_DEPTH_TEST);
	} else {
		glDisable(GL_DEPTH_TEST);
	}
	glDepthFunc(GL_LEQUAL);

	


	// Clear the background
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Set matrix pipeline to identity
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	/* Setup the projection matrix. */
	gluPerspective(window->fovy, (GLfloat) window->width / (GLfloat)window->height,
	               window->pnear, window->pfar);

	/* Now, post-multiply the projection matrix with the camera matrix */
	gluLookAt(camera->eye.x, camera->eye.y, camera->eye.z,
	          camera->center().x, camera->center().y, camera->center().z,
	          camera->up.x, camera->up.y, camera->up.z);

	

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Set the clear color to black
	glClearColor(0.0, 0.0, 0.0, 0.0);

	// Set the shading model
	glShadeModel(GL_SMOOTH);
	//glShadeModel(GL_FLAT);

	// Set the polygon mode to fill
	glPolygonMode(GL_FRONT_AND_BACK, fillPolygons ? GL_FILL : GL_LINE);


	// Define material properties of specular color and degree of 
	// shininess.  Since this is only done once in this particular 
	// example, it applies to all objects.  Material properties can 
	// be set for individual objects, individual faces of the objects,
	// individual vertices of the faces, etc... 

	GLfloat specular [] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat shininess [] = { 100.0 };

	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

	// Set the GL_AMBIENT_AND_DIFFUSE color state variable to be the
	// one referred to by all following calls to glColor
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_NORMALIZE);

	// Create a Directional Light Source
	GLfloat position [] = { 5.0, 10.0, 5.0, 0.0 };
	glLightfv(GL_LIGHT0, GL_POSITION, position);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);


	

	
	glColor3f(.75, .5, .25);

	terrain->draw(camera, locked);

	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	
	gluQuadricOrientation(quad, GLU_OUTSIDE);

	list<Particle *> * newList = new list<Particle *>();

	bool listChanged = false;


	for(list<Particle *>::iterator vli = particleList->begin(); vli != particleList->end(); vli++) {
		Particle * p = *vli;
		p->draw();
		p->timeStep();
		if(p!=NULL && p->validPos()) {
			newList->push_back(p);
		} else {
			if(p!=NULL) {
				int k1 = p->key1();
				int k2 = p->key2();

				hash_map<int, Particle*> * box1 = (*ht1)[k1];
				if(box1!=NULL) {
					box1->erase(p->key());
				}

				hash_map<int, Particle*> * box2 = (*ht2)[k2];
				if(box2!=NULL) {
					box2->erase(p->key());
				}
			}
			

			delete p;
			listChanged=true;
		}
	}
	delete particleList;
	particleList = newList;
	if(listChanged) {
		//cout << "Balls: " << particleList->size() << " Reason: " << reason << "\n";
		
	}

	glutSwapBuffers();
}


/* This function is called whenever the window is resized.
 * Parameters are the new dimentions of the window
 */
static void resize(int x, int y) {
	glViewport(0, 0, x, y);
	
	window->width = x;
	window->height = y;

	glutPostRedisplay();
}


/* This function is called whenever the mouse is pressed or released
 * button is a number 0 to 2 designating the button
 * state is 1 for release 0 for press event
 * x and y are the location of the mouse (in window-relative coordinates)
 */
static void mouseButton(int button, int state, int x, int y) {
	
	lastx = x;
	lasty = y;
	
}


/* This function is called whenever the mouse is moved with a mouse button held
 * down. x and y are the location of the mouse (in window-relative coordinates)
 */
static void mouseMotion(int x, int y) {
	GLint dx = x - lastx; lastx = x;
	GLint dy = y - lasty; lasty = y;

	GLfloat thetaX = dx/100.0;
	GLfloat thetaY = dy/100.0;

	*camera = (camera->rotateX(thetaY)).rotateY(-thetaX);
	*trans = RotateYMatrix(-thetaX) * (*trans);
	glutPostRedisplay();
}

static void idle() {
	
	if(rain) {
		GLfloat x, z;

		for(int i = 0; i < 32; i++) {
			GLfloat alpha = ((GLfloat) rand())/((GLfloat)RAND_MAX);
			GLfloat beta = ((GLfloat) rand())/((GLfloat)RAND_MAX)*(1-alpha);

			Vector2D dropXZ = terrainTransform * Vector2D(alpha, beta);

			x = dropXZ.x/10;
			z = (dropXZ.z - 7)/10;

			if(x==0 || z==0) {
				cout << "x: " << x << " z: " << z << "\n";
				cout << "alpha: " << beta << " beta: " << alpha << "\n";
			}

			particleList->push_back(new Particle(Vector(x,2,z,1), Vector(0,0,0,1), terrain, quad, ht1, ht2));
		}
	}
	if(trans->x * trans->x + trans->y * trans->y + trans->z * trans->z > EPSILON || particleList->size()>0) {
		glutPostRedisplay();
	}
}


/* This function is called whenever there is a keyboard input
 * key is the ASCII value of the key pressed
 * x and y are the location of the mouse
 */
static void keyDown(unsigned char key, int x, int y) {
	Vector moveVector;

	
	switch(key) {
	case 0x1B: /* ESC */

		printf("Exiting.\n");
		exit(0);
		break;
	case 'r':
	case 'R':
		*trans=Vector();
		resetCamera();
		glutPostRedisplay();
		break;
	case 'l':
	case 'L':
		locked = !locked;
		printf("Locking %s.\n", locked ? "on" : "off");
		break;
	case 'Z':
	case 'z':
		depthTest = !depthTest;
		printf("Z-Buffering is now %s.\n", fillPolygons ? "on" : "off");
		glutPostRedisplay();
		break;

	case 'W':
	case 'w':
		moveVector = (Vector(camera->look.x, 0, camera->look.z,0)).normalize();
		*trans = *trans + KB_INTERVAL * moveVector;
	
		wDown=true;
		glutPostRedisplay();
		break;
	case 'S':
	case 's':
		moveVector = (Vector(camera->look.x, 0, camera->look.z,0)).normalize();
		*trans = *trans - KB_INTERVAL * moveVector;
		glutPostRedisplay();
		sDown=true;
		break;
	case 'A':
	case 'a':
		moveVector = (camera->up % (Vector(camera->look.x, 0, camera->look.z,0))).normalize();
		*trans = *trans + KB_INTERVAL * moveVector;
		aDown=true;
		glutPostRedisplay();
		break;
	case 'D':
	case 'd':
		moveVector = (camera->up % (Vector(camera->look.x, 0, camera->look.z,0))).normalize();
		*trans = *trans - KB_INTERVAL * moveVector;
		dDown=true;
		glutPostRedisplay();
		break;
	case 'E':
	case 'e':
		*trans = *trans + Vector(0, KB_INTERVAL, 0, 0);
		eDown=true;
		glutPostRedisplay();
		break;
	case 'Q':
	case 'q':
		*trans = *trans + Vector(0, -KB_INTERVAL, 0, 0);
		qDown=true;
		glutPostRedisplay();
		break;



	case 'f':
	case 'F':
		fillPolygons = !fillPolygons;
		printf("Polygons are now %s.\n", fillPolygons ? "filled" : "outlined");
		glutPostRedisplay();
		break;

	case 32: /* SPACEBAR */
		rain = true;
		
		glutPostRedisplay();
		break;

	default:
		fprintf(stderr, "Unrecognized key: %d(%c)\n", (int)key, key);
		break;
	}
}

/* This function is called whenever there is a keyboard input
 * key is the ASCII value of the key pressed
 * x and y are the location of the mouse
 */
static void keyUp(unsigned char key, int x, int y) {
	
	Vector moveVector;
	switch(key) {
	
	case 'W':
	case 'w':
		moveVector = (Vector(camera->look.x, 0, camera->look.z,0)).normalize();
		*trans = *trans - KB_INTERVAL * moveVector;
		wDown=false;
		killMotion();
		break;
	case 'S':
	case 's':
		moveVector = (Vector(camera->look.x, 0, camera->look.z,0)).normalize();
		*trans = *trans + KB_INTERVAL * moveVector;
		sDown=false;
		killMotion();
		break;
	case 'A':
	case 'a':
		moveVector = (camera->up % (Vector(camera->look.x, 0, camera->look.z,0))).normalize();
		*trans = *trans - KB_INTERVAL * moveVector;
		aDown=false;
		killMotion();
		break;
	case 'D':
	case 'd':
		moveVector = (camera->up % (Vector(camera->look.x, 0, camera->look.z,0))).normalize();
		*trans = *trans + KB_INTERVAL * moveVector;
		dDown=false;
		killMotion();
		break;
	case 'E':
	case 'e':
		*trans = *trans - Vector(0, KB_INTERVAL, 0, 0);
		eDown=false;
		killMotion();
		break;
	case 'Q':
	case 'q':
		*trans = *trans - Vector(0, -KB_INTERVAL, 0, 0);
		qDown=false;
		killMotion();
		break;


	case 32: // SPACEBAR 
		rain = false;
		cout << "Rain" <<"\n";
		glutPostRedisplay();
		break;

	default:
		break;

	}
}



/* Here's the main */
int main(int argc, char* argv[]) {

	srand( (unsigned)time( NULL ) );


	
	/* Allocate space for our data */
	window = (WINDOW *)calloc(1, sizeof(WINDOW));
	window->width = RES_X;
	window->height = RES_Y;
	window->pnear = PNEAR;
	window->pfar = PFAR;
	window->fovy = FOVY;

	/* Initialize GLUT */
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

	glutCreateWindow("CS184 Assignment 4");
	glutReshapeWindow(window->width, window->height);

	glutDisplayFunc(display);
	glutReshapeFunc(resize);
	glutMouseFunc(mouseButton);
	glutMotionFunc(mouseMotion);
	glutKeyboardFunc(keyDown);
	glutKeyboardUpFunc(keyUp);
	//glutSpecialFunc(keydown);
	//glutSpecialUpFunc(keyup);
	glutIdleFunc(idle);
	glutIgnoreKeyRepeat(1); 


	resetCamera();
	

	terrain = new LandscapeTriangle(NULL, new Vertex(new Vector( 0, 0, -7, 0)),
										  new Vertex(new Vector( 8, 0,  7, 0)),
										  new Vertex(new Vector(-8, 0,  7, 0)), 0, 16, true);


	Vector2D a = Vector2D( 8, 14);
	Vector2D b = Vector2D(-8, 14);

	terrainTransform = Matrix2D(a, b);


	// Switch to main loop
	glutMainLoop();
	return 0;
}

