#include <iostream>
#include "LandscapeTriangle.h"
#include "Vector.h"
#include "Vertex.h"
#include "Camera.h"
#include "Matrix2D.h"
#include "Vector2D.h"



const GLfloat SUBDIVIDE_THRESHOLD = .01;
const int MAX_SUBDIVISIONS = 8;
//const GLfloat LN2R = 1.4426950408889634073599246810019;




LandscapeTriangle::LandscapeTriangle(LandscapeTriangle * parent, Vertex * a, Vertex * b, Vertex * c, int level, GLfloat sideLength, bool orientedUp) {
	
	this->normal = NULL;
	
	this->parent=parent;
	this->corner[0]=a;
	this->corner[1]=b;
	this->corner[2]=c;
	this->level=level;
	this->isSubdivided=false;
	this->sideLength=sideLength;
	this->neighbor[0]=this->neighbor[1]=this->neighbor[2]=NULL;
	this->orientedUp=orientedUp;
	this->wasDivided = false;

	this->normal=new Vector();
	*(this->normal) = (*(c->p)-*(a->p)) % (*(b->p)-*(a->p));

	if(orientedUp) {
		for(int i = 0; i<3; i++) {
			corner[i]->set_up(i, this);
		}
	} else {
		for(int i = 0; i < 3; i++) {
			corner[i]->set_down(i, this);
		}
	}
}

void LandscapeTriangle::draw(Camera * camera, bool locked) {

	if((locked && wasDivided) || (!locked && closeTo(camera))) {
		
		if(!this->isSubdivided) {
			this->subdivide(locked, this);
		}
		wasDivided = true;
		for(int i = 0; i < 3; i++) {
			//setColor(i+1);
			outside[i]->draw(camera, locked);
		}
		//setColor(3);
		center->draw(camera, locked);
		//setColor(4);

	} else {
		wasDivided = false;
		int sd[3] = {-1, -1, -1};	//index of 1st, 2nd, and 3rd subdivided neighbors
		int sdn = 0;		//# subdivided neighbors;
		bool sdnc = false;	//a neighbor's child is subdivided

		LandscapeTriangle * ltns[3];


		
		if(!locked) {
			//check whether neighbors should be subdivided -- if so, make sure they are
			for(int i = 0; i < 3; i++) {
				if(neighbor[i] != NULL) {					
					if(neighbor[i]->closeTo(camera)) {
						
						ltns[sdn] = neighbor[i];

						if(!neighbor[i]->isSubdivided) {
							neighbor[i]->subdivide(locked, this);
						}

						for(int j = 0; j < 2; j++) {
							int n = index(i+j);
							if(neighbor[i]->outside[n]->closeTo(camera)) {
								if(!neighbor[i]->outside[n]->isSubdivided) {
									neighbor[i]->outside[n]->subdivide(locked, this);
								}
								sdnc = true;
							}
						}
						sd[sdn] = i;
					
						sdn++;
				
					}
				}
			}
		}

		for(int i = 0; i<sdn; i++) {
			if(ltns[i]->isSubdivided==false) {
				printf("Error1: Neighbor %d not subdivided.\n", i);
			}
			if(neighbor[sd[i]]->isSubdivided==false) {
				printf("Error2: Neighbor %d not subdivided.\n", i);
			}
		}

		int i0, i1, i2, i3;
		
		switch(sdn) {
			case 0:
				glBegin(GL_TRIANGLES);
					for(int i = 0; i < 3; i++) {
						corner[i]->draw();
					}
				glEnd();
				break;

			case 1:
				i0 = sd[0]; //i=index of first subdivided neighbor
				
				glBegin(GL_TRIANGLE_STRIP);
					corner[i0]->draw();
					neighbor[i0]->edge[i0]->draw();
					corner[index(i0+2)]->draw();
					corner[index(i0+1)]->draw();
				glEnd();
				break;
			case 2:
				
				if(sd[0] == index(sd[1]+1)) {
					i1 = sd[1];
					i2 = sd[0];
				} else {
					i1 = sd[0];
					i2 = sd[1];
				}
				i3=index(i2+1);
				
				glBegin(GL_TRIANGLE_FAN);
					neighbor[i1]->edge[i1]->draw();
					corner[i2]->draw();
					neighbor[i2]->edge[i2]->draw();
					corner[i3]->draw();
					corner[i1]->draw();
				glEnd();
				break;
				//break;
			default: //2 or 3
				this->subdivide(locked, this);
				for(int i = 0; i < 3; i++) {
					outside[i]->draw(camera, locked);
				}
				center->draw(camera, locked);
				break;
		}

	}
}

void LandscapeTriangle::subdivideNeighbors(LandscapeTriangle * caller) {
	
	if(!this->isSubdivided) {
		this->subdivide(false, caller);
	}
	for(int i = 0; i < 3; i++) {
		if(neighbor[i] != NULL && !neighbor[i]->isSubdivided) {
			neighbor[i]->subdivide(false, caller);
		}
	}
}

void LandscapeTriangle::subdivide(bool locked, LandscapeTriangle * caller) {

	if(this->parent != NULL && !locked) {
		this->parent->subdivideNeighbors(caller);
	}
	GLfloat multiplier = this->sideLength * this->getRange();
	
	//create edge vertices
	GLfloat displacement;
	for(int i = 0; i<3; i++) {
		if(neighbor[i] != NULL && neighbor[i]->isSubdivided) {
			edge[i] = neighbor[i]->edge[i];
		} else {
			displacement = multiplier * ((rand()/(RAND_MAX/2.0))-1);
			edge[i] = new Vertex(new Vector(	(corner[i]->p->x + corner[index(i+1)]->p->x)/2,
												(corner[i]->p->y + corner[index(i+1)]->p->y)/2 + displacement,
												(corner[i]->p->z + corner[index(i+1)]->p->z)/2,
												1));
		}
	}

	int nLevel = level+1;
	GLfloat nSideLength = 0.5 * sideLength;
	
	//create subdivided triangles
	//for(int i = 0; i < 3; i++) {
		//outside[i] = new LandscapeTriangle(this, corner[i], edge[i], edge[index(i+2)], nLevel, nSideLength, orientedUp);
	//}
	outside[0] = new LandscapeTriangle(this, corner[0], edge[0], edge[2], nLevel, nSideLength, orientedUp);
	outside[1] = new LandscapeTriangle(this, edge[0], corner[1], edge[1], nLevel, nSideLength, orientedUp);
	outside[2] = new LandscapeTriangle(this, edge[2], edge[1], corner[2], nLevel, nSideLength, orientedUp);
	center = new LandscapeTriangle(this, edge[1], edge[2], edge[0], nLevel, nSideLength, !orientedUp);
	
	//assign internal neighbor relations
	for(int i = 0; i < 3; i++) {
		outside[i]->neighbor[index(i+1)] = center;
		center->neighbor[index(i+1)] = outside[i];
	}

	//assign external neighbor relations
	for(int i = 0; i < 3; i++) {
		if(neighbor[i]!=NULL && neighbor[i]->isSubdivided) {
			outside[i]->neighbor[i] = neighbor[i]->outside[index(i+1)];
			neighbor[i]->outside[index(i+1)]->neighbor[i] = outside[i];
			
			outside[index(i+1)]->neighbor[i] = neighbor[i]->outside[i];
			neighbor[i]->outside[i]->neighbor[i] = outside[index(i+1)];
		}
	}

	
	this->isSubdivided=true;
}

GLfloat LandscapeTriangle::getRange() {
	switch(level) {
		case 0: return 0;
		case 1: return 0;
		case 2: return 0;
		case 3: return .25;
		default: return 1.0/(level-2);
	}
}

bool LandscapeTriangle::closeTo(Camera * camera) {


	if(level > MAX_SUBDIVISIONS) {
		return false;
	}

	Vector eyeTo[3];

	Vector m = (((*(corner[0]->p)) + (*(corner[1]->p))) + (*(corner[2]->p)))/3;
	
	//Vector eye2m = m - camera->eye;
	for(int i = 0; i < 3; i++) {
		eyeTo[i] = *(corner[i]->p) - camera->eye;
	}
	
	if((eyeTo[0] * camera->look <=0) && (eyeTo[1] * camera->look <=0) && (eyeTo[2] * camera->look <=0)) {
		return false;
	}

	GLfloat dTo[3];
	for(int i = 0; i < 3; i++) {
		dTo[i] = eyeTo[i] * eyeTo[i];
	}
	return (this->sideLength / dTo[0] >= SUBDIVIDE_THRESHOLD)
		|| (this->sideLength / dTo[1] >= SUBDIVIDE_THRESHOLD)
		|| (this->sideLength / dTo[2] >= SUBDIVIDE_THRESHOLD);
}


LandscapeTriangle::~LandscapeTriangle(void) {
	delete [] (LandscapeTriangle **) outside;
	delete center;

	for(int i = 0; i < 3; i++) {
		if(neighbor[i] == NULL) {
			delete edge[i];
		} else {
			neighbor[i]->neighbor[i]=NULL;
			if(edge[i] != NULL) {
				if(this->orientedUp) {
					edge[i]->unset_up(i);
				} else {
					edge[i]->unset_down(i);
				}
			}
		}
	}
}

inline int index(int n) {
	return n % 3;
}

void setColor(int c) {
	switch(c) {
		case 0: glColor3f(0,0,0); break;
		case 1: glColor3f(0,0,1); break;
		case 2: glColor3f(0,1,0); break;
		case 3: glColor3f(1,0,0); break;
		case 4: glColor3f(1,1,0); break;
		case 5: glColor3f(0,1,1); break;
		case 6: glColor3f(1,0,1); break;
		case 7: glColor3f(.1, .5, .25); break;
		case 8: glColor3f(.9, .9, .9); break;
		case 9: glColor3f(.2, .2, .3); break;
		default: glColor3f(.75, .5, .25); break;
	}
}


bool LandscapeTriangle::hit(GLfloat x, GLfloat z, GLfloat * height, Vector ** normal) {
	Vector2D v = Vector2D(x - corner[0]->p->x, z - corner[0]->p->z);

	Vector2D a = Vector2D(corner[1]->p->x - corner[0]->p->x, corner[1]->p->z - corner[0]->p->z);
	Vector2D b = Vector2D(corner[2]->p->x - corner[0]->p->x, corner[2]->p->z - corner[0]->p->z);

	Matrix2D A = Matrix2D(a, b);
	Vector2D xv = A.inverse()*v;
	
	if((xv.x >= 0) && (xv.z >= 0) && (xv.x + xv.x <= 1)) {
		*height = this->getY(xv.x, xv.z, normal);
		//cout << "input: " << v << ", HIT: " << xv << "\n";
		//cout << "a = " << a << ", b = " << b <<"\n";
		return true;
	} else {
		//cout << "input: " << v << ", MISS: " << xv << "\n";
		//cout << "a = " << a << ", b = " << b <<"\n";
		*normal = NULL;
		return false;
	}
}

GLfloat LandscapeTriangle::getY(GLfloat alpha, GLfloat beta, Vector ** normal) {
	if(!this->isSubdivided) {
		*normal = this->normal;
		return (1-alpha-beta)*corner[0]->p->y + alpha * corner[1]->p->y + beta * corner[2]->p->y;
	} else {
		if((alpha < 0 || beta < 0 || (alpha + beta) > 1)) {
			return false;
		} else if(alpha > .5) {
			return outside[1]->getY(2*alpha - 1, 2*beta, normal);
		} else if(beta  > .5) {
			return outside[2]->getY(2*alpha, 2*beta - 1, normal);
		} else if(alpha + beta < .5) {
			return outside[0]->getY(2*alpha, 2*beta, normal);
		} else {
			return center->getY(1-2*alpha, 1-2*beta, normal);
		}
	}
}