#include "Matrix.h"
#include "Vector.h"




Matrix::Matrix() {
	this->matrix = new GLfloat[rows*cols];
}

Matrix ::~Matrix(void) {
	//delete [] this->matrix;
}


Matrix Matrix ::transpose(void) const {
	Matrix M = Matrix();
	for(unsigned int i = 1; i <= rows; i++) {
		for (unsigned int j = 1; j <= cols; j++) {
			M.set(j, i, get(i, j));
		}
	}
	return M;
}
/*
Matrix Matrix ::inverse(void) const {
	Matrix C = coFactorMatrix();
	Matrix T = C.transpose();
	Matrix I = 1.0/determinant() * T;
	return I;
	
}
*/
/*
GLfloat Matrix ::determinant(void) const {
	
}
*/
/*
Matrix Matrix::coFactorMatrix() {
	Matrix M = Matrix(rows, cols);
	for(unsigned int i = 1; i <= rows; i++) {
		for(unsigned int j = 1; j <= cols; j++) {
			GLfloat sign = -1;
			if(((i+j) % 2) == 0) {
				sign = 1;
			}
			Matrix & S = this->subMatrix(i, j);
			GLfloat d = S.determinant();
			delete & S;
			GLfloat c = sign * d;
			M.set(i, j, c);
		}
	}
	return M;
}
*/

Matrix operator* (Matrix & A, Matrix & B) {
	Matrix M = Matrix();
	for (unsigned int i = 1; i <= rows; i ++) {
		for (unsigned int j = 1; j <= cols; j ++) {
			GLfloat x = 0;
			for(unsigned int k = 1; k <=cols; k++) {
				x += A.get(i,k) * B.get(k, j);
			}
			M.set(i, j, x);
		}
	}
	return M;
}



Matrix operator+ (Matrix & A, Matrix & B) {
	Matrix M = Matrix();
	for (unsigned int i = 1; i <= rows; i++) {
		for (unsigned int j = 1; j <= cols; j++) {
			M.set(i, j, A.get(i, j) + B.get(i, j));
		}
	}
	return M;
}

Matrix operator- (Matrix  & A, Matrix  & B) {
	Matrix M = Matrix();
	for (unsigned int i = 1; i <= rows; i++) {
		for (unsigned int j = 1; j <= cols; j++) {
			M.set(i, j, A.get(i, j) - B.get(i, j));
		}
	}
	return M;
}

Matrix operator* ( Matrix  & A,  GLfloat k) {
	Matrix M = Matrix();
	for (unsigned int i = 1; i <= rows; i++) {
		for (unsigned int j = 1; j <= cols; j++) {
			M.set(i, j, k * A.get(i, j));
		}
	}
	return M;
}

Matrix operator* (GLfloat k,  Matrix  & A) {
	return A * k;
}

Vector operator* (Matrix & A, Vector & v) {
	GLfloat x = v.x * A.get(1, 1) + v.y * A.get(1, 2) + v.z * A.get(1, 3) + v.w * A.get(1, 4);
	GLfloat y = v.x * A.get(2, 1) + v.y * A.get(2, 2) + v.z * A.get(2, 3) + v.w * A.get(2, 4);
	GLfloat z = v.x * A.get(3, 1) + v.y * A.get(3, 2) + v.z * A.get(3, 3) + v.w * A.get(3, 4);
	GLfloat w = v.x * A.get(4, 1) + v.y * A.get(4, 2) + v.z * A.get(4, 3) + v.w * A.get(4, 4);
	return Vector(x, y, z, w);
}


ostream& operator<< (ostream& os, Matrix& M) {
	os << '\n';
	for(unsigned int row = 1; row <= rows; row++) {
		os << "| ";
		for(unsigned int col = 1; col <= cols; col++) {
			os << M.get(row, col) << ' ';
		}
		os << "|\n";
	}
	os << '\n';
	return os;
}

Matrix RotateXMatrix(GLfloat theta) {
	Matrix M = Matrix();
	M.set(1, 1, 1); M.set(1, 2, 0         ); M.set(1, 3,  0         ); M.set( 1, 4, 0);
	M.set(2, 1, 0); M.set(2, 2, cos(theta)); M.set(2, 3, -sin(theta)); M.set( 2, 4, 0);
	M.set(3, 1, 0); M.set(3, 2, sin(theta)); M.set(3, 3,  cos(theta)); M.set( 3, 4, 0);
	M.set(4, 1, 0); M.set(4, 2, 0         ); M.set(4, 3,  0         ); M.set( 4, 4, 1);
	return M;
}

Matrix RotateYMatrix(GLfloat theta) {
	Matrix M = Matrix();
	M.set(1, 1,  cos(theta)); M.set(1, 2, 0); M.set(1, 3,  sin(theta)); M.set( 1, 4, 0);
	M.set(2, 1,  0         ); M.set(2, 2, 1); M.set(2, 3,  0         ); M.set( 2, 4, 0);
	M.set(3, 1, -sin(theta)); M.set(3, 2, 0); M.set(3, 3,  cos(theta)); M.set( 3, 4, 0);
	M.set(4, 1,  0         ); M.set(4, 2, 0); M.set(4, 3,  0         ); M.set( 4, 4, 1);
	return M;	
}

Matrix RotateZMatrix(GLfloat theta) {
	Matrix M = Matrix();
	M.set(1, 1,  cos(theta)); M.set(1, 2, -sin(theta)); M.set(1, 3,  0); M.set( 1, 4, 0);
	M.set(2, 1,  sin(theta)); M.set(2, 2,  cos(theta)); M.set(2, 3,  0); M.set( 2, 4, 0);
	M.set(3, 1,  0         ); M.set(3, 2,  0         ); M.set(3, 3,  1); M.set( 3, 4, 0);
	M.set(4, 1,  0         ); M.set(4, 2,  0         ); M.set(4, 3,  0); M.set( 4, 4, 1);
	return M;	
}

Matrix TranslateMatrix(GLfloat x, GLfloat y, GLfloat z) {
	Matrix M = Matrix();
	M.set(1, 1, 1); M.set(1, 2, 0); M.set(1, 3, 0); M.set(1, 4, x);
	M.set(2, 1, 0); M.set(2, 2, 1); M.set(2, 3, 0); M.set(2, 4, y);
	M.set(3, 1, 0); M.set(3, 2, 0); M.set(3, 3, 1); M.set(3, 4, z);
	M.set(4, 1, 0); M.set(4, 2, 0); M.set(4, 3, 0); M.set(4, 4, 1);
	return M;
}

Matrix TranslateMatrix(const Vector & v) {
	Matrix M = Matrix();
	M.set(1, 1, 1); M.set(1, 2, 0); M.set(1, 3, 0); M.set(1, 4, v.x);
	M.set(2, 1, 0); M.set(2, 2, 1); M.set(2, 3, 0); M.set(2, 4, v.y);
	M.set(3, 1, 0); M.set(3, 2, 0); M.set(3, 3, 1); M.set(3, 4, v.z);
	M.set(4, 1, 0); M.set(4, 2, 0); M.set(4, 3, 0); M.set(4, 4, 1);
	return M;
}