#ifndef MATRIXH
#define MATRIXH

#include <stdlib.h>
#include <iostream>

#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

class Vector;

class Matrix {
public:
	Matrix(void);
	
	friend Matrix operator* ( Matrix & A,  Matrix & B);
	friend Matrix operator+ ( Matrix & A,  Matrix & B);
	friend Matrix operator- ( Matrix & A,  Matrix & B);
	friend Matrix operator* ( Matrix & A,  GLfloat k);
	friend Matrix operator* ( GLfloat k,  Matrix & A);
	//friend Vector operator* ( Vector & v, Matrix & A);
	friend Vector operator* ( Matrix & A, Vector & v);

	inline void set(unsigned int row, unsigned int col, GLfloat value);
	inline GLfloat get(unsigned int row, unsigned int col) const ;
	Matrix transpose(void) const;
	//Matrix inverse(void) const;
	//Matrix subMatrix(unsigned int row, unsigned int col) const;
	//Matrix coFactorMatrix() const;
	//GLfloat determinant(void) const;
	~Matrix(void);

private:
	GLfloat * matrix;
};

const static unsigned int rows = 4;
const static unsigned int cols = 4;


ostream & operator<< (ostream& os, Matrix& M);

inline void Matrix::set(unsigned int row, unsigned int col, GLfloat value) {
	matrix[ (row-1) * cols + col - 1] = value;

}

inline GLfloat Matrix::get(unsigned int row, unsigned int col) const {
	return matrix[ (row-1) * cols + col - 1];
}

Matrix RotateXMatrix(GLfloat theta);
Matrix RotateYMatrix(GLfloat theta);
Matrix RotateZMatrix(GLfloat theta);
Matrix TranslateMatrix(GLfloat x, GLfloat y, GLfloat z);
Matrix TranslateMatrix(const Vector & v);

#endif


