#include <math.h>

#include "Vector.h"
#include "Particle.h"
#include "LandscapeTriangle.h"
//#include "IntVector.h"

const static GLfloat DT = .05;
const static GLfloat G = 9.81;
const static GLfloat E = .15;
const static GLfloat FM = 20;
const static GLfloat R = .04;
const static GLfloat RESTING_DISTANCE = .01;	//number of radii between resting points
const static GLfloat MIN_D = .0001;
const static GLfloat DM = 1/RESTING_DISTANCE;
const static GLfloat F = .25;	//friction
const static GLint BOX_WIDTH = .04;
const static GLint MAX_ITER = 3;
const static GLfloat MAX_FORCE = 20;

Particle::Particle(Vector pos, Vector vel, LandscapeTriangle * terrain, GLUquadric * quad, hash_map<int, hash_map<int, Particle*> * > * ht1, hash_map<int, hash_map<int, Particle*> * > * ht2) {

	this->pos = pos;
	this->vel = vel;
	this->terrain = terrain;
	this->quad = quad;
	this->valid = true;
	this->ht1 = ht1;
	this->ht2 = ht2;
	
	this->iv1=this->getIV1();
	this->iv2=this->getIV2();

	hash_map<int, Particle*> * box1 = (*ht1)[this->key1()];
	hash_map<int, Particle*> * box2 = (*ht2)[this->key2()];

	if(box1 == NULL) {
		box1 = new hash_map<int, Particle*>;
		(*ht1)[this->key1()]=box1;
	}

	if(box2 == NULL) {
		box2 = new hash_map<int, Particle*>;
		(*ht2)[this->key2()]=box2;
	}

	(*box1)[this->key()] = this;
	(*box2)[this->key()] = this;
}

Particle::~Particle(void) {
}

void Particle::draw() {
	glPushMatrix();
		GLfloat speed = vel.magnitude() / 2;
		speed = speed>1 ? 1 : speed;
		glColor4f(speed, speed, .5 + speed/2 , .25);
		glTranslatef(pos.x, pos.y, pos.z);
		gluSphere(quad, R, 6, 6);
	glPopMatrix();
}

void Particle::timeStep() {
	GLfloat height;
	Vector extraNormal=Vector(0,1,0,0);
	Vector * normal = &extraNormal;
	valid=terrain->hit(pos.x, pos.z, &height, &normal);
	//Vector oldPos = pos;
	
	if(!valid) {
		return;
	}
	
	IntVector iv;
	hash_map<int, hash_map<int, Particle*> *> ht;
	int key;

	Vector oldVel = vel;
	Vector sum = Vector(0, -G, 0, 0);

	if(this->BoxDistance1() <= this->BoxDistance2()) {
		iv = iv1;
		key = key1();
		ht = *ht1;
	} else {
		iv = iv2;
		key = key2();
		ht = *ht2;
	}


	

	hash_map<int, Particle*> * box = ht[key];
	if(box != NULL) {
		for(hash_map<int, Particle*>::iterator boxi = box->begin(); boxi != box->end(); boxi++) {
			Particle * p = boxi->second;
			if(p!=NULL && p!= this && p->valid) {
				Vector delta = (p->pos - this->pos);
				GLfloat distance = delta.magnitude();
				if(distance > 0) {
					GLfloat force = emforce(distance);

					
					Vector dn = delta.normalize();

					sum = sum + dn * force;

				}

			}
		}
	}

	

	Vector newPos = pos + vel*DT;



	vel = vel + DT*sum;
	GLfloat dy = newPos.y-height;
	
	int count = 0;

	bool vTemp = valid;
	


	while(vTemp && dy < 0 && vel.y < 0 && count < MAX_ITER && valid) {

		
		
		//Vector drop = Vector(0, dy, 0, 0);
		//Vector along = drop.projectionOnTo(*normal)
		//Vector i1	= pos - along;
		//pos = pos - 2 * (drop - drop.projectionOnTo(*normal));	//should take into account energy loss, but doesn't
		//if(vel.y < 0) {
		

		Vector proj = vel.projectionOnTo(*normal);
		vel = (vel-proj)*(1-proj.magnitude()*F) - E * proj;
		//}
		newPos = pos + vel*DT;
		vTemp=terrain->hit(pos.x, pos.z, &height, &normal);
		GLfloat dy = newPos.y-height;
		count++;
		if(count == MAX_ITER) {
			newPos.y=height;
			vel.y=0;
		}
	}



	pos = newPos;



	if(this->iv1!=this->getIV1()) {
		hash_map<int, Particle*> * box1 = (*ht1)[this->key1()];
		if(box1!=NULL) {
			box1->erase(this->key());
		}
		
		this->iv1=this->getIV1();
		box1 = (*ht1)[this->key1()];

		if(box1==NULL) {
			box1 = new hash_map<int, Particle*>();
			(*ht1)[this->key1()] = box1;
		}
		(*box1)[this->key()]=this;
	}


	
	if(this->iv2!=this->getIV2()) {
		hash_map<int, Particle*> * box2 = (*ht2)[this->key2()];
		if(box2!=NULL) {
			box2->erase(this->key());
		}
		
		this->iv2=this->getIV2();
		box2 = (*ht2)[this->key2()];

		if(box2==NULL) {
			box2 = new hash_map<int, Particle*>();
			(*ht2)[this->key2()] = box2;
		}
		(*box2)[this->key()]=this;
	}
	

	
}

bool Particle::validPos() {
	return this->valid;
}

GLfloat emforce(GLfloat distance) {
	if(distance < MIN_D) distance = MIN_D;
	GLfloat ddm = distance * DM;
	GLfloat ddm2 = ddm*ddm;
	GLfloat force = FM * (1-ddm2)/(ddm2*ddm2);
	return force>MAX_FORCE ? MAX_FORCE : force;
}

IntVector Particle::getIV1() const {
	return IntVector(	(GLint) (pos.x / BOX_WIDTH),
						(GLint) (pos.y / BOX_WIDTH),
						(GLint) (pos.z / BOX_WIDTH)   );
}

IntVector Particle::getIV2() const {
	return IntVector(	(GLint) (0.5 + pos.x / BOX_WIDTH),
						(GLint) (0.5 + pos.y / BOX_WIDTH),
						(GLint) (0.5 + pos.z / BOX_WIDTH)   );
}

GLfloat Particle::BoxDistance1() const {
	Vector c = Vector(	BOX_WIDTH * iv1.x,
						BOX_WIDTH * iv1.y,
						BOX_WIDTH * iv1.z,
						1);
	Vector d = c - this->pos;
	return d.magnitude();
}

GLfloat Particle::BoxDistance2() const {
	Vector c = Vector(	BOX_WIDTH * (iv2.x - .5),
						BOX_WIDTH * (iv2.y - .5),
						BOX_WIDTH * (iv2.z - .5),
						1);
	Vector d = c - this->pos;
	return d.magnitude();
}


int Particle::key() const {
	return (int) this;
}


bool invalid(Vector * v) {
	return !(v->x > -30 && v->y > -30 && v->z > -30 && v->x < 30 && v->y < 30 && v->z < 30);
}