#ifndef VECTORH
#define VECTORH

#include <stdlib.h>
#include <math.h>
#include <iostream>


#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>

//#include "Matrix.h"
//#include "vectormath.h"

using namespace std;

class Vector {
public:
	Vector(void);
//	Vector(VECTOR * v);
	Vector(GLfloat x, GLfloat y, GLfloat z, GLfloat w);
	//Vector & operator=(const Vector & u);

	friend Vector    operator+  (const Vector & u,  const Vector & v);
	friend Vector    operator-  (const Vector & u,  const Vector & v);
	friend GLfloat   operator*  (const Vector & u,  const Vector & v); //dot product
	friend Vector    operator%  (const Vector & u,  const Vector & v); //cross product
	friend Vector    operator*  (const Vector & u,  const GLfloat  k);
	friend Vector    operator*  (      GLfloat  k,  const Vector & u);
	friend Vector    operator/  (const Vector & v,        GLfloat  k);
	friend bool      operator== (const Vector & u,  const Vector & v);
	friend ostream & operator<< (ostream      & os, const Vector & v);

	
	~Vector(void);

//	VECTOR * toVECTOR(void) const;

	GLfloat magnitude(void) const;
	Vector normalize(void) const;
	//Matrix  crossProductMatrix(void);
	Vector projectionOnTo(const Vector /*&*/ v) const;
//private:
	GLfloat x;
	GLfloat y;
	GLfloat z;
	GLfloat w;
};


inline GLfloat Vector::magnitude(void) const {
	return sqrt(x*x + y*y + z*z);	
}

inline Vector Vector::normalize(void) const {
	GLfloat m = this->magnitude();
	return Vector(x / m, y / m, z / m, 0);
}
	
inline Vector operator+(const Vector & u, const Vector & v){
	return Vector(u.x + v.x, u.y + v.y, u.z + v.z, (GLfloat) ((int) u.w | (int) v.w));
}
inline Vector operator-(const Vector & u, const Vector & v){
	return Vector(u.x - v.x, u.y - v.y, u.z - v.z, u.w);
}
inline GLfloat operator*(const Vector & u, const Vector & v){ //dot product
	return (u.x * v.x) + (u.y * v.y) + (u.z * v.z);
}
inline Vector operator%(const Vector & u, const Vector & v){ //cross product
	return Vector(u.y*v.z - u.z*v.y, u.z*v.x - u.x*v.z, u.x*v.y-u.y*v.x, 0);
}

inline Vector operator*(const Vector & u, GLfloat  k){ 
	return Vector(u.x * k, u.y * k, u.z *k, 0);
}

inline Vector operator*(GLfloat k, const Vector & u){ 
	return Vector(u.x * k, u.y * k, u.z *k, 0);
}

inline Vector operator/(const Vector & u, GLfloat k){
	return Vector(u.x / k, u.y / k, u.z / k, 0);
}


inline bool operator==(const Vector & u, const Vector & v){
	return (u.x==v.x) && (u.y==v.y) && (u.z==v.z) && (u.w==v.w);
}

inline ostream& operator<< (ostream& os, const Vector& v) {
	os << '(' <<  v.x << ", " << v.y << ", " << v.z << ')';
	return os;
}

inline Vector Vector::projectionOnTo(const Vector /*&*/ v) const {
	GLfloat vv = v*v;
	if(vv != 0) {
		return ((v * (*this))/(vv)) * v;
	} else {
		return Vector(0,0,0,0);
	}
}


//Vector  operator@(Vector  u, Vector  v); //angle

#endif

