#include "Vertex.h"
#include "Vector.h"
#include "LandscapeTriangle.h"

Vertex::Vertex(Vector * p) {
	this->p = p;
	
	for(int i = 0; i < 3; i++) {
		up[i] = NULL;
		down[i] = NULL;
		up_flag[i] = 0;
		down_flag[i] = 0;
	}

	flagSum = 0;

	normal = new Vector(0,0,0,0);
}

Vertex::~Vertex(void) {
	delete normal;
}

Vector* Vertex::getNormal(void) {
	if(this->sum() == flagSum) {
		return this->normal;
	} else {
		flagSum = sum();
		*normal = Vector(0,0,0,0);
		for(int i = 0; i<3; i++) {
			if(up_flag[i] == 1) {
				*normal = *normal + up[i]->normal->normalize();
			}
			if(down_flag[i] == 1) {
				*normal = *normal + down[i]->normal->normalize();
			}
		}

		return normal;
	}
}


void Vertex::set_up(int n, LandscapeTriangle * lt) {
	this->up[n] = lt;
	this->up_flag[n] = 1;
}

void Vertex::set_down(int n, LandscapeTriangle * lt) {
	this->down[n] = lt;
	this->down_flag[n] = 1;
}
//////////////
void Vertex::unset_up(int n) {
	this->up[n] = NULL;
	this->up_flag[n] = 0;
}

void Vertex::unset_down(int n) {
	this->down[n] = NULL;
	this->down_flag[n] = 0;
}

void Vertex::draw() {
	if(p->y > .6) {
		setColor(8);
	} else if (p->y >.5) {
		setColor(9);
	} else if (p->y > 0) {
		setColor(7);
	} else {
		setColor(10);
	}
	Vector * n = this->getNormal();
	glNormal3f(n->x, n->y, n->z);
	glVertex3f(p->x, p->y, p->z);
}

inline int Vertex::sum() {
	return up_flag[0] + up_flag[1] + up_flag[2]
		+ down_flag[0] + down_flag[1] + down_flag[2];
}